import cupy as cp


class Frame:
    def __init__(self, mask):
        """A container to the mask with useful features."""
        self.mask = mask

    def get_lbls(self):
        return cp.unique(self.mask)

    def is_empty(self):
        """
        return if the frame is empty.
        """
        return len(self.get_lbls()) == 1

    def get_locations(self):
        """
        returns the centroids of each cell in the frame.
        """
        lbls = self.get_lbls()[1:]
        locations = []
        # compute the average
        for lbl in lbls:
            coords = cp.asarray((self.mask == lbl)).T  # mask to coord
            locations.append(cp.average(coords, axis=0))
        return cp.array(locations)



import cupy as cp
import numpy as np
import torch
from torch.utils.data import DataLoader
from cellstitch_cuda.seg_batch import ImageDataset
import sys
from cupyx.scipy.ndimage import zoom


def downscale_mask(masks, pixel=None, z_res=None):
    if not pixel:
        pixel = 1
    if not z_res:
        z_res = 1

    masks = cp.asarray(masks)

    anisotropy = z_res / pixel
    zoom_factors = (1 / (pixel / 0.5), 1 / (anisotropy * (pixel / 0.5)), 1)
    order = 0  # 0 nearest neighbor, 1 bilinear, 2 quadratic, 3 bicubic

    masks = zoom(masks, zoom_factors, order=order).get()
    cp._default_memory_pool.free_all_blocks()

    return masks


def upscale_img(images, pixel=None, z_res=None):
    if not pixel:
        pixel = 1
    if not z_res:
        z_res = 1

    anisotropy = z_res / pixel
    zoom_factors = (
        pixel / 0.5,
        anisotropy * (pixel / 0.5),
        1,
    )  # Pre-zoom for InstanSeg (expected pixel size = 0.5)
    order = 1  # 0 nearest neighbor, 1 bilinear, 2 quadratic, 3 bicubic

    zoomed = []
    for ch in images:  # CiZk
        ch = zoom(cp.asarray(ch), zoom_factors, order=order).get()
        cp._default_memory_pool.free_all_blocks()
        zoomed.append(ch)

    cp._default_memory_pool.free_all_blocks()

    return np.stack(zoomed)


def histogram_correct(images, match: str = "first"):
    """Correct bleaching over a given axis

    This function is used to correct signal degradation that can occur over the Z axis.

    Adapted from napari-bleach-correct: https://github.com/marx-alex/napari-bleach-correct
        Authored by https://github.com/marx-alex
        Original algorithm by Kota Miura: Miura K. Bleach correction ImageJ plugin for compensating the photobleaching
        of time-lapse sequences. F1000Res. 2020 Dec 21;9:1494. https://doi.org/10.12688/f1000research.27171.1

    """
    # cache image dtype
    dtype = images.dtype

    assert (
        3 <= len(images.shape) <= 4
    ), f"Expected 3d or 4d image stack, instead got {len(images.shape)} dimensions"

    avail_match_methods = ["first", "neighbor"]
    assert (
        match in avail_match_methods
    ), f"'match' expected to be one of {avail_match_methods}, instead got {match}"

    images = images.transpose(1, 0, 2, 3)  # ZCYX --> CZYX

    corrected = []
    for ch in images:
        ch = _correct(cp.asarray(ch), match).get()
        cp._default_memory_pool.free_all_blocks()
        corrected.append(ch)

    images = np.stack(corrected, axis=1, dtype=dtype)  # ZCYX

    return images


def _correct(channel, match):

    # channel = cp.array(channel)
    k, m, n = channel.shape
    pixel_size = m * n

    # flatten the last dimensions and calculate normalized cdf
    channel = channel.reshape(k, -1)
    values, cdfs = [], []

    for i in range(k):

        if i > 0:
            if match == "first":
                match_ix = 0
            else:
                match_ix = i - 1

            val, ix, cnt = cp.unique(
                channel[i, ...].flatten(), return_inverse=True, return_counts=True
            )
            cdf = cp.cumsum(cnt) / pixel_size

            interpolated = cp.interp(cdf, cdfs[match_ix], values[match_ix])
            channel[i, ...] = interpolated[ix]

        if i == 0 or match == "neighbor":
            val, cnt = cp.unique(channel[i, ...].flatten(), return_counts=True)
            cdf = cp.cumsum(cnt) / pixel_size
            values.append(val)
            cdfs.append(cdf)

    channel = channel.reshape(k, m, n)

    return channel


def segment_single_slice_medium(
    d,
    model,
    tiles,
    batch_size,
):
    res = model.eval_medium_image(
        d,
        return_image_tensor=False,
        pixel_size=None,
        target="all_outputs",
        cleanup_fragments=True,
        tile_size=tiles,
        batch_size=batch_size,
        normalise=True,
    )
    return res[0]


def segment_single_slice_small(d, model):
    res = model.eval_small_image(
        d,
        return_image_tensor=False,
        pixel_size=None,
        target="all_outputs",
        cleanup_fragments=True,
        normalise=True,
    )
    return res[0]


def segment_batch_slice_small(d, model):
    result = []
    for batch in d:
        batch = batch.to(model.inference_device)
        target_segmentation = torch.tensor([1, 1])
        with torch.amp.autocast("cuda"):
            instanseg_kwargs = {"cleanup_fragments": True}
            instances = model.instanseg(
                batch, target_segmentation=target_segmentation, **instanseg_kwargs
            )
        res = instances.cpu()
        result.append(res.numpy().astype("uint32"))
    result = [
        k for b in result for k in b
    ]  # For each z plane found in each batch b in result, stack the z planes
    result = np.stack(result).transpose(
        1, 2, 3, 0
    )  # ZcYX --> cYXZ || kcij --> cijk (for iterator k)

    return result[1], result[0]


def segmentation(d, model, m: str = "nuclei_cells", xy: bool = False):

    mode = 1  # Base for 'nuclei_cells' and 'cells'
    if m == "nuclei":
        mode = 0

    nuclei_cells = False
    if xy and m == "nuclei_cells":
        nuclei_cells = True

    nslices = d.shape[-1]

    vram = torch.cuda.mem_get_info()[0] / 1024  # In KB
    vram_est = 0.1765 * np.prod(
        d.shape[0:3]
    )  # Magic number literally obtained by plotting in Excel

    tiles = 1024
    if vram < vram_est:
        small = False
        vram_est = (
            0.1765 * tiles**2 * d.shape[0]
        )  # Base VRAM estimate on batch size, multiplied by channels
        batch = int(vram / vram_est)
        if batch == 0:
            print(
                "Not enough VRAM available for 1024x1024 tiles. Decreasing to standard 512x512."
            )
            tiles = 512
            vram_est = 0.1765 * tiles**2 * d.shape[0]
            batch = int(vram / vram_est)
            if batch == 0:
                print("Not enough VRAM available for 512x512 tiles. Aborting.")
                sys.exit(1)
    else:
        small = True

    if small:  # For images that fit within VRAM in their entirety:
        batch = int(vram / vram_est)
        if batch > 3:  # If over 3 slices would fit into VRAM in their entirety
            d = d.transpose(3, 0, 1, 2)  # CYXZ --> ZCYX || Cijk --> kCij
            dataset = ImageDataset(
                d
            )  # Create a dataset that handles images similar to Instanseg
            dataloader = DataLoader(
                dataset, batch_size=batch - 2, shuffle=False, drop_last=False
            )  # Leverage torch batching through DataLoader
            empty_res, nuclei = segment_batch_slice_small(dataloader, model)
        else:  # Otherwise, go slice by slice
            empty_res = np.zeros_like(d[0])
            nuclei = empty_res.copy()
            for xyz in range(nslices):
                res_slice = segment_single_slice_small(d[:, :, :, xyz], model)
                empty_res[:, :, xyz] = res_slice[mode]
                if nuclei_cells:
                    nuclei[:, :, xyz] = res_slice[0]
    else:  # For larger images
        empty_res = np.zeros_like(d[0])
        nuclei = empty_res.copy()
        for xyz in range(nslices):
            res_slice = segment_single_slice_medium(
                d[:, :, :, xyz], model, tiles, batch
            )
            empty_res[:, :, xyz] = res_slice[mode]
            if nuclei_cells:
                nuclei[:, :, xyz] = res_slice[0]
    if nuclei_cells:
        return empty_res, nuclei
    return empty_res



from torch.utils.data import Dataset
import torch
import numpy as np
from instanseg.utils.utils import percentile_normalize


class ImageDataset(Dataset):
    def __init__(self, image_list):
        self.image_list = image_list

    def __len__(self):
        return len(self.image_list)

    def __getitem__(self, idx):

        image = self.image_list[idx]

        if isinstance(image, np.ndarray):
            if image.dtype == np.uint16:
                image = image.astype(np.int32)
            image = torch.from_numpy(image).float()

        image = image.squeeze()

        assert (
            3 >= image.dim() >= 2
        ), f"Input image shape {image.shape} is not supported."

        image = torch.atleast_3d(image)

        image = percentile_normalize(image)

        return image



import ot
import ot.plot
import numpy as np
import cupy as cp
import fill_voids
import time

from cellpose import utils as cp_utils
from joblib import Parallel, delayed
from skimage.measure import regionprops


# -------------------------------
# interpolation helper functions
# -------------------------------


def comp_match_plan(pc1, pc2, dist="sqeuclidean"):
    """Compute optimal matching plans between 2 sets of point clouds"""
    # compute cost matrix
    C = ot.dist(pc1, pc2, metric=dist).astype(np.float64)
    C /= C.max()

    # convert point clouds to uniform distributions
    n_pts1, n_pts2 = pc1.shape[0], pc2.shape[0]
    mu1, mu2 = np.ones(n_pts1) / n_pts1, np.ones(n_pts2) / n_pts2

    # compute transport plan
    plan = ot.emd(mu1, mu2, C)

    return plan


def interpolate(pc1, pc2, dist="sqeuclidean", anisotropy=2):
    """
    Calculate interpolated predictions

    Parameters
    ----------
    pc1 : np.ndarray
        Point cloud representing cell boundary in frame 1

    pc2 : np.ndarray
        Point cloud representing cell boundary in frame 2

    anisotropy : int
        Ratio of sampling rate between different axes

    Returns
    -------
    interp_pcs : list
        Smoothed boundary locations along interpolated layers
    """
    alphas = np.linspace(0, 1, anisotropy + 1)[1:-1]
    plan = comp_match_plan(pc1, pc2, dist=dist)
    normalized_plan = plan / plan.sum(
        axis=1, keepdims=1
    )  # normalize so that the row sum is 1

    interp_pcs = []

    for alpha in alphas:
        n_pts = pc1.shape[0]
        avg_pc = np.zeros((n_pts, 2), dtype=int)

        for i in range(n_pts):
            point = pc1[i]
            target_weights = normalized_plan[i]

            weighted_target = np.array(
                [np.sum(target_weights * pc2[:, 0]), np.sum(target_weights * pc2[:, 1])]
            )

            avg_pc[i, :] = point * (1 - alpha) + alpha * weighted_target

        interp_pcs.append(avg_pc)

    return interp_pcs


# -----------------------------------------
# Util functions for interp reconstruction
# -----------------------------------------


def get_lbls(mask):
    """Get unique labels from the predicted masks"""
    return np.unique(mask[mask != 0])


def min_size_filter(res, thld=100):
    """Filter out all masks with area below threshold"""
    assert len(res) == 4, "Invalid Cellpose 4-tuple result"
    preds = res[0]

    for i in range(len(preds)):
        lbls = get_lbls(preds[i])
        for lbl in lbls:
            msk = preds[i] == lbl
            if msk.sum() < thld:
                coords = np.nonzero(msk)
                preds[i][coords] = 0

    res_filtered = (preds, res[1], res[2], res[3])
    return res_filtered


def get_contours(masks):
    """Transfer solid mask predictions to non-overlapping contours w/ distinct integers"""
    masks_new = masks.copy()
    outlines = cp_utils.masks_to_outlines(masks)
    masks_new[~outlines] = 0

    return masks_new


def get_mask_perimeter(masks, lbl, is_contour=False):
    assert lbl in get_lbls(
        masks
    ), "Label {} doesn't in current mask predictions".format(lbl)

    if is_contour:
        p = (masks == lbl).sum()
    else:
        mask_lbl = (masks == lbl).astype(np.uint8)
        p = cp_utils.masks_to_outlines(mask_lbl).sum()

    return p


def calc_vols(pred):
    """
    Calculate volumes of each mask from predictions
    """
    lbls = get_lbls(pred)
    vols = [(pred == lbl).sum() for lbl in lbls]
    return vols


def calc_depth(masks):
    """
    Calculate z-layer depth of predictions
    """
    assert masks.ndim == 3, "Mask predictions must be 3D to calculate depth"

    lbls = get_lbls(masks)
    depths = np.vectorize(lambda lbl: np.diff(np.nonzero(masks == lbl)[0][[0, -1]])[0])(
        lbls
    )

    return depths


def mask_to_coord(mask):
    """Return (n, 2) coordinates from masks"""
    coord = np.asarray(np.nonzero(mask)).T
    return coord


def coord_to_mask(coord, size, lbl):
    """Convert from coordinates to original labeled masks"""
    mask = np.zeros(size)
    mask[tuple(coord.T)] = lbl
    return mask


def contour_to_mask(contour):
    lbl = get_lbls(contour)[0]
    """ Convert contour to solid masks with fill-in labels"""
    binary_contour = contour > 0
    binary_mask = fill_voids.fill(np.asarray(binary_contour))

    mask = np.zeros_like(binary_contour)
    mask[binary_mask] = lbl

    return mask


def connect(coord1, coord2, mask):
    """
    Modify the mask by connecting the two given coordinates.

    Parameters
    ----------
    coord1 : [x1, y1]
        Coordinate of the first pixel.

    coord2 : [x2, y2]
        Coordinate of the second pixel.

    mask : binary np.narray
        Binary mask of the boundary.
    """
    x1, y1 = coord1
    x2, y2 = coord2

    x_offset, y_offset = x2 - x1, y2 - y1

    # skip if the two coordinates are already connected
    if x_offset**2 + y_offset**2 <= 2:
        return

    # Determine the number of steps to connect the points
    steps = max(abs(x_offset), abs(y_offset))
    x_values = np.linspace(x1, x2, steps + 1, dtype=np.int32)
    y_values = np.linspace(y1, y2, steps + 1, dtype=np.int32)

    mask[x_values, y_values] = 1


def calc_angles(source_point, target_points, eps=1e-20):
    """
    Calculate angle (rad) between source point (source_point) & list of target points(target_points, dim=(n, 2))
    """
    source_points = np.tile(source_point, (target_points.shape[0], 1))
    diff = target_points - source_points
    angles = np.apply_along_axis(
        lambda x: np.arctan2(x[1], x[0] + eps), axis=1, arr=diff
    )

    return angles


def connect_boundary(coords, size, lbl=1):
    """
    Connect interpolation coordinates to generate close-loop mask contours

    Parameters
    ----------
    coords : ndarray, shape (ns,2)
        Boundary coordinates (might be disconnected).

    size: (n1, n2)
        Shape of the final mask.

    lbl: int
        Label of the original mask.

    Returns
    -------
    mask: ndarray, shape (n1, n2)
        Connected boundary mask.

    """
    # Sort boundary labels by angle to mask's mass center
    mass_center = np.round(coords.mean(0)).astype(np.int64)
    angles = calc_angles(mass_center, coords)
    sorted_coords = coords[angles.argsort()]

    mask = coord_to_mask(coords, size, lbl)

    for i, (x, y) in enumerate(sorted_coords[:-1]):
        next_x, next_y = sorted_coords[i + 1]
        connect((x, y), (next_x, next_y), mask)

    connect(tuple(sorted_coords[-1]), tuple(sorted_coords[0]), mask)

    return mask


def get_mask_center(mask):
    coordx = np.round(np.nonzero(mask)[0].sum() / mask.sum()).astype(np.uint16)
    coordy = np.round(np.nonzero(mask)[1].sum() / mask.sum()).astype(np.uint16)

    return coordx, coordy


def get_mask_center_cupy(mask):

    if cp.sum(mask) == 0:
        return cp.zeros(2, dtype=cp.uint16)

    # Get indices and calculate centroids
    x_indices = cp.nonzero(mask)[0]
    y_indices = cp.nonzero(mask)[1]
    count = cp.sum(mask)

    coordx = cp.round(x_indices.sum() / count).astype(cp.uint16)
    coordy = cp.round(y_indices.sum() / count).astype(cp.uint16)

    return cp.array([coordx, coordy])


# -----------------------------
# Core Interpolation functions
# -----------------------------


def process_region(label, label_slice, cell_mask, dist, anisotropy):

    coordinates = label_slice

    source_ct = cell_mask[0].astype(np.uint8)
    target_ct = cell_mask[1].astype(np.uint8)

    shape = source_ct.shape

    source_coord = mask_to_coord(source_ct)
    target_coord = mask_to_coord(target_ct)

    interp_coords = interpolate(
        source_coord, target_coord, dist=dist, anisotropy=anisotropy
    )

    interps = [
        fill_voids.fill(connect_boundary(interp, shape)) * label
        for interp in interp_coords
    ]

    return interps, list(coordinates)


def interp_layers(source_target, dist="sqeuclidean", anisotropy=2, n_jobs=-1):
    """
    Interpolating adjacent z-layers using bounding boxes after contouring
    """

    interp_masks = np.zeros(
        (
            anisotropy + 1,  # num. interpolated layers
            source_target.shape[1],  # y
            source_target.shape[2],  # x
        ),
        dtype=source_target.dtype,
    )

    def _dilation(coords, lims):
        y, x = coords
        ymax, xmax = lims
        dy, dx = np.meshgrid(
            np.arange(y - 2, y + 3), np.arange(x - 2, x + 3), indexing="ij"
        )
        dy, dx = dy.flatten(), dx.flatten()
        mask = np.logical_and(
            np.logical_and(dy >= 0, dx >= 0), np.logical_and(dy < ymax, dx < xmax)
        )
        return dy[mask], dx[mask]

    source_mask = source_target[0]
    target_mask = source_target[1]

    shape = source_mask.shape
    source_contour = get_contours(source_mask)
    target_contour = get_contours(target_mask)

    # Boundary condition: if empty on source / target label
    # align the empty slice w/ mass centers to represent instance endings
    source_dummy = np.zeros_like(source_mask)
    target_dummy = np.zeros_like(target_mask)
    if not np.intersect1d(get_lbls(source_contour), get_lbls(target_contour)).size:
        if (source_contour.sum() == target_contour.sum() == 0) or (
            np.logical_and(source_mask, target_mask).sum() > 0
        ):
            return np.zeros(shape)
        for lbl in get_lbls(source_contour):
            yc, xc = _dilation(get_mask_center(source_mask == lbl), source_mask.shape)
            target_dummy[yc, xc] = lbl
        for lbl in get_lbls(target_contour):
            yc, xc = _dilation(get_mask_center(target_mask == lbl), target_mask.shape)
            source_dummy[yc, xc] = lbl
        source_contour += source_dummy
        target_contour += target_dummy

    # Determine which labels are intersecting
    joint_lbls = np.intersect1d(get_lbls(source_contour), get_lbls(target_contour))

    # Reassign source_target
    source_target = np.stack((source_contour, target_contour))

    # Filter source_target for labels that are found to intersect
    source_target = np.where(np.isin(source_target, joint_lbls), source_target, 0)

    # Use regionprops to set up bounding boxes based on the newly generated contours
    regions = regionprops(source_target)

    # Parallel process bounding boxes
    results = Parallel(n_jobs=n_jobs)(
        delayed(process_region)(
            region.label,
            region.slice,
            region.image,
            dist,
            anisotropy,
        )
        for region in regions
    )

    masks, coordinates = zip(*results)

    # Write each interpolated bounding box into the empty interp_masks array
    for i, mask in enumerate(masks):
        z_min, z_max = 1, anisotropy
        y_min, y_max = coordinates[i][1].start, coordinates[i][1].stop
        x_min, x_max = coordinates[i][2].start, coordinates[i][2].stop

        interp_masks[z_min:z_max, y_min:y_max, x_min:x_max] = np.where(
            interp_masks[z_min:z_max, y_min:y_max, x_min:x_max] == 0,
            # interp_masks[z_min:z_max, y_min:y_max, x_min:x_max] < mask,  # Should behave similar to picking max value
            mask,
            interp_masks[z_min:z_max, y_min:y_max, x_min:x_max],
        )

    interp_masks[0] = source_mask
    interp_masks[-1] = target_mask

    return interp_masks


def full_interpolate(masks, anisotropy=2, dist="sqeuclidean", n_jobs=-1, verbose=False):
    """
    Interpolating between all adjacent z-layers

    Parameters
    ----------
    masks : np.ndarray
        layers of 2D predictions
        (dim: (Depth, H, W))

    anisotropy : int
        Ratio of sampling rate between xy-axes & z-axis

    Returns
    -------|
    interp_masks : np.ndarray
        interpolated masks
        (dim: (Depth * anisotropy - (anisotropy-1), H, W))

    """
    if masks.max() < 256:
        masks = masks.astype("uint8")
    elif masks.max() < 65536:
        masks = masks.astype("uint16")

    dtype = masks.dtype

    interp_masks = np.zeros(
        (
            len(masks) + (len(masks) - 1) * (anisotropy - 1),
            masks.shape[1],
            masks.shape[2],
        ),
        dtype=dtype,
    )

    idx = 0
    for i in range(masks.shape[0] - 1):
        if verbose:
            time_start = time.time()
            print("Interpolating layer {} & {}...".format(i, i + 1))
        source_target = masks[i : i + 2].copy()
        interps = interp_layers(
            source_target, dist=dist, anisotropy=anisotropy, n_jobs=n_jobs
        )
        interp_masks[idx : idx + anisotropy + 1] = interps
        idx += anisotropy
        if verbose:
            print("Time to interpolate: ", time.time() - time_start)

    return interp_masks



import tifffile
from cellstitch_cuda.pipeline import cellstitch_cuda
from cellstitch_cuda.interpolate import full_interpolate

img = r"E:\1_DATA\Rheenen\tvl_jr\SP8\2024Nov22_SI_1mg_1year_3D_2D-merge\Villus_crops\2024Nov22_SI_1mg_1year_3D_2D-merge-1-1\output.tif"

cellstitch_cuda(img, output_masks=True, verbose=True, seg_mode="nuclei_cells", interpolation=True, n_jobs=-1)

# masks = tifffile.imread(r"E:\1_DATA\Rheenen\tvl_jr\3d stitching\raw4-small\cellstitch_masks.tif")
#
# masks = full_interpolate(masks, anisotropy=4, n_jobs=-1, verbose=True)
#
# tifffile.imwrite(r"E:\1_DATA\Rheenen\tvl_jr\3d stitching\raw4-small\cellstitch_masks_interpolated.tif", masks)
